#!/usr/bin/env perl

use strict;
use warnings;
use v5.38.0;

use Path::Tiny;
use Linux::ACL;
use File::ExtAttr;
use Data::Dumper;

use Getopt::Long;
use builtin qw/true false/;
no warnings 'experimental::builtin';

my $o_verbose = false;
my $o_dryrun  = false;
my $o_copyfilerange = false;

my $o_hash_alg = "SHA-256"; # TODO implement this

GetOptions(
  "v|verbose" => sub {$o_verbose = true},
  "n|dryrun"  => sub {$o_dryrun  = true},
  "f|full"    => sub {$o_hash_alg = "full"},
  "copyfilerange" => sub {$o_copyfilerange = true},
  "h|help"    => \&display_help,
);

sub display_help() {
  print <DATA>;
  exit();
}

if (@ARGV != 2) {
  my $n = scalar @ARGV;
  print STDERR "Incorrect number of file arguments.  Got $n, expected 2\n\n";
  display_help();
}

sub get_facl($file) {
}

sub get_times($file) {
}

sub get_xattr($file) {
  my %xattr;

  my @namespaces = File::ExtAttr::listfattrns($file);

  for my $ns (@namespaces) {
    my @xkeys = File::ExtAttr::listfattr($file, {namespace => $ns});
    for my $key (@xkeys) {
      my $value = File::ExtAttr::getfattr($file, $key, {namespace => $ns});
      $xattr{$ns}{$key} = $value;
    }
  }

  print Dumper(\%xattr);
}

my $file = path($ARGV[0]);
print "Reading $file\n";
get_xattr($file);

__DATA__
block-cloning-tool
------------------

Tool for helping offline dedup for OpenZFS 2.2+ using the new block-cloning feature

The intent is to preserve the file permissions, file owner, m-time and a-time, and even
the file ACLs and extended-attributes so that you don't lose anything like you would
with a normal `cp` reflink command.

Right now this only works on linux, it will need tweaks on non-Linux systems due to
the use of the Linux::ACL library.  Any pointers at what can do similar on other systems
is welcome (I'm aware of Solaris::ACL)

Requirements
============
Enable the block-cloning feature on the pool:

  zpool set feature@block_cloning=enabled poolname

And then you'll want to set the zfs parameter zfs_bclone_enabled to 1.  I recommend doing 
this by creating a /etc/modprobe.d/zfs.conf and adding this line.

  options zfs zfs_bclone_enabled=1

You will then need to regenerate your initrd which will vary by linux distro but on debian
based systems it will look like the following:

  update-initramfs -k all -u

Usage
=====
   -h --help          This help text

   -n --dryrun        Perform a dry run, this will perform all checks and output if it
                      would perform the ioctl() call or not.

   -f --full          Perform a complete byte by byte check of the files rather than
                      using a hashing algorithm.  Not sure this will ever be necessary
                      but I'm adding it for anyone who is paranoid.

      --copyfilerange Use the copyfilerange() syscall, not enabled by default as there
                      are a number of caveats about using it.  It will always succeed,
                      even if block cloning was not possible.  It tells the Linux
                      kernel to make the copy in any way that is possible, even if it's
                      a direct, normal copy of the file

   -v --verbose       Output the information about the ACLs, xattr and file times that
                      are collected about the destination file so that the information
                      can be logged and reviewed.


